// Copyright 2024 The Tessera authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package main

import (
	"testing"
)

func TestLeafGenerator(t *testing.T) {
	// Always generate new values
	gN := newLeafGenerator(0, 100, 0)
	vs := make(map[string]bool)
	for range 256 {
		v := string(gN())
		vs[v] = true
	}

	// Always generate duplicate
	gD := newLeafGenerator(256, 100, 1.0)
	for range 256 {
		if !vs[string(gD())] {
			t.Error("Expected duplicate")
		}
	}
}
